import cli2
import litellm
import os

from prompt2.backend import BackendPlugin


class LiteLLMBackend(BackendPlugin):
    def __init__(self, model_name, **model_kwargs):
        self.model_name = model_name
        self.model_kwargs = model_kwargs

    async def completion(self, messages):
        if os.getenv('DEBUG'):
            litellm._turn_on_debug()
        stream = await litellm.acompletion(
            messages=messages,
            stream=True,
            model=self.model_name,
            **self.model_kwargs,
        )

        full_content = ""
        async for chunk in stream:
            if hasattr(chunk, 'choices') and chunk.choices:
                delta = chunk.choices[0].delta
                if hasattr(delta, 'content') and delta.content is not None:
                    full_content += delta.content

        return full_content



import cli2
import importlib.metadata

from .exception import NotFoundError


class Parser:
    entry_point = 'prompt2_parser'

    class NotFoundError(NotFoundError):
        title = 'PARSER NOT FOUND'

        def available_list(self):
            plugins = importlib.metadata.entry_points(
                group=Parser.entry_point,
            )
            return [p.name for p in plugins]

    def __init__(self, model):
        self.model = model

    @classmethod
    def get(cls, name):
        plugins = importlib.metadata.entry_points(
            name=name,
            group=Parser.entry_point,
        )
        if not plugins:
            raise cls.NotFoundError(name)
        return [*plugins][0].load()

    def messages(self, messages):
        return messages

    def parse(self, response):
        return response


class Wholefile(Parser):
    system = "Respond only with the complete new file content, as this will be directly written to the file by an automated AI assistant tool—no additional text, comments, or explanations are allowed."  # noqa

    def messages(self, messages):
        messages.append(
            dict(
                role='system',
                content=self.system,
            ),
        )
        return messages

    def parse(self, response):
        if response.startswith('```'):
            # strip markup the IA absolutely wants to add
            return '\n'.join([l for l in response.split('\n')[1:-1]])
        return response



"""
Prompt management.
"""
import cli2
import functools
import importlib.metadata
from pathlib import Path
import os
import re

import jinja2

from .exception import NotFoundError
from .model import Model


class Prompt:
    """
    Build prompts from text files and context variables.
    """
    if 'PROMPT2_USER_PATH' in os.environ:
        USER_PATH = Path(os.getenv('PROMPT2_USER_PATH'))
    else:
        USER_PATH = Path(os.getenv('HOME')) / '.prompt2/prompts'

    if 'PROMPT2_LOCAL_PATH' in os.environ:
        LOCAL_PATH = Path(os.getenv('PROMPT2_LOCAL_PATH'))
    else:
        LOCAL_PATH = Path(os.getcwd()) / '.prompt2/prompts'

    entry_point = 'prompt2_paths'

    class NotFoundError(NotFoundError):
        pass

    def __init__(self, *prompts, **context):
        self.parts = []
        for prompt in prompts:
            self.read(prompt)
        self.context = context

    @classmethod
    def default_paths(cls, _cli2=None):
        paths = [cls.LOCAL_PATH]
        if cls.USER_PATH != cls.LOCAL_PATH:
            paths.append(cls.USER_PATH)
        if _cli2:
            return [str(p) for p in paths]
        return paths

    @classmethod
    def paths(cls):
        plugins = importlib.metadata.entry_points(
            group=cls.entry_point,
        )
        paths = []
        for plugin in plugins:
            paths += plugin.load()()
        return paths

    @classmethod
    def find(cls, name, paths=None):
        paths = paths or cls.paths()
        for path in paths:
            path = path / f'{name}.txt'
            if path.exists():
                return path
        raise cls.NotFoundError(name, paths)

    @classmethod
    def names(cls, paths=None):
        """ List prompt names """
        paths = paths or cls.default_paths()
        names = set()
        for path in self.paths():
            if not path.exists():
                continue
            for file in path.iterdir():
                names.add(file.name[:-4])
        return list(names)

    def read(self, name):
        """
        Read a prompt text file by name.

        :param name: Name of the prompt, as returned by names
        """
        path = self.find(name)
        with path.open('r') as f:
            content = f.read()
            cli2.log.debug(
                'prompt loaded',
                path=str(path),
                content=content,
            )
            self.parts.append(content)
            self.path = path
            return content

    def render(self):
        # collect all paths from path plugins ...
        plugins = importlib.metadata.entry_points(
            group='prompt2_paths',
        )
        paths = [
            plugin.load()() for plugin in plugins
        ]

        # ... to build a jinja2 environment ...
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                [str(p) for p in paths],
            ),
            undefined=jinja2.StrictUndefined,
            autoescape=False,
        )

        # ... in which you can hook your plugins
        plugins = importlib.metadata.entry_points(
            group='prompt2_globals',
        )
        for plugin in plugins:
            env.globals[plugin.name] = plugin.load()

        # Render the template with the data
        template = env.from_string('\n'.join(self.parts))
        return template.render(**self.context)

    def messages(self):
        messages = [
            dict(
                role='user',
                content=self.render(),
            ),
        ]
        return messages



class Prompt2Error(Exception):
    title = 'PROMPT2 ERROR'


class NotFoundError(Exception):
    title = 'NOT FOUND'

    def __init__(self, name, searched=None):
        self.name = name
        self.searched = [str(s) for s in searched or []]
        msg = [f'{self.title}']

        if self.searched:
            msg += ['SEARCHED:'] + searched

        try:
            self.available = self.available_list()
        except NotImplementedError:
            self.available = None
        else:
            msg += ['AVAILABLE:'] + self.available

        super().__init__(msg)

    def available_list(self):
        raise NotImplementedError()



"""
Functions that will be exposed in Jinja2.

Add yours over the code2_jinja2 entry point plugin!
"""
import prompt2


def read(path):
    with open(path, 'r') as f:
        content = f.read()
    return content



from setuptools import setup


setup(
    name='prompt2',
    versioning='dev',
    setup_requires='setupmeta',
    packages=['prompt2'],
    install_requires=[
        'cli2',
        'litellm',
        'markdown-it-py',
        'jinja2',
    ],
    author='James Pic',
    author_email='jamespic@gmail.com',
    url='https://yourlabs.io/oss/cli2',
    include_package_data=True,
    license='MIT',
    keywords='cli',
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'prompt2 = prompt2.cli:cli.entry_point',
        ],
        'prompt2_parser': [
            'wholefile = prompt2.parser:Wholefile',
        ],
        'prompt2_globals': [
            'read = prompt2.jinja2:read',
        ],
        'prompt2_paths': [
            'paths = prompt2.prompt:Prompt.default_paths',
        ],
        'prompt2_backend': [
            'litellm = prompt2.backends.litellm:LiteLLMBackend',
        ],
        'pytest11': [
            'prompt2 = prompt2.pytest',
        ],
    },
)



"""
Template based prompts on the CLI
"""

import cli2
import functools
import importlib
import inspect
import jinja2
import os
from pathlib import Path

import prompt2


class PromptCommand(cli2.Command):
    def __init__(self, *args, **kwargs):
        self._model = None
        self._parser = None
        self._prompt = None
        self.not_found_error = None
        super().__init__(*args, **kwargs)

    def parse(self, *argv):
        super().parse(*argv)

        try:
            # auto cast variables dependending on each other ...
            if 'model' in self:
                if self['model'].value:
                    self['model'].value = prompt2.Model.get(
                        self['model'].value,
                        strict=True,
                    )
                else:
                    self['model'].value = prompt2.Model.get()

            if 'parser' in self and self['parser'].value:
                self['parser'].value = (
                    self['model'].value.parser(self['parser'].value)
                )

            if 'prompt' in self:
                context = dict()

                try:
                    context = self['context'].value
                except (cli2.Cli2ValueError, KeyError):
                    pass

                self['prompt'].value = prompt2.Prompt(
                    self['prompt'].value,
                    **context,
                )

        except prompt2.NotFoundError as exc:
            # we'll deal with that during actual call
            self.not_found_error = exc
        except cli2.Cli2ValueError:
            pass

    def call(self, *args, **kwargs):
        if self.not_found_error:
            exc = self.not_found_error
            print(cli2.t.red.bold(exc.title) + f': {exc.name}\n')
            if exc.available:
                print(cli2.t.green.bold('AVAILABLE') + ':')
                return exc.available
            elif exc.searched:
                print(cli2.t.green.bold('SEARCHED') + ':')
                return exc.searched
        return super().call(*args, **kwargs)

    def handle_exception(self, exc):
        if isinstance(exc, jinja2.UndefinedError):
            print(cli2.t.red.bold(f'MISSING CONTEXT VARIABLE'))
            print(exc.message)
        else:
            raise exc


cli = cli2.Group(
    doc=__doc__,
    cmdclass=PromptCommand,
)


@cli.cmd(color='green')
def paths():
    """
    Return prompt paths
    """
    return [str(p) for p in prompt2.Prompt.paths()]


@cli.cmd(color='green', name='list')
def prompts():
    """
    List available prompts
    """
    paths = dict()
    for path in prompt2.Prompt.paths():
        if not path.exists():
            continue
        paths.update({
            p.name[:-4]: str(p)
            for p in path.iterdir()
        })
    return paths


@cli.cmd
def edit(name, local: bool=False):
    """
    Edit a prompt.

    :param name: Prompt name.
    :param local: Enable this to store in $CWD/.prompt2 instead of
                  $HOME/.prompt2
    """
    local_path = prompt2.Prompt.LOCAL_PATH / f'{name}.txt'
    if local_path.exists() or local:
        path = local_path
    else:
        user_path = prompt2.Prompt.USER_PATH / f'{name}.txt'
        if user_path.exists():
            path = user_path
        else:
            path = local_path if local else user_path

    if path.exists():
        with path.open('r') as f:
            content = f.read()
    else:
        content = (
            'You are called by an automated process, you MUST'
            ' structure your reply or you will crash it!'
        )

    content = cli2.editor(content)
    if content:
        path.parent.mkdir(exist_ok=True, parents=True)
        with path.open('w') as f:
            f.write(content)
        cli2.log.info('wrote', path=str(path), content=content)

        print(cli2.t.bold('SAVED PROMPT:'))
        print(content + '\n')
        print(cli2.t.green(f'Saved to {path}'))


@cli.cmd(color='green')
def show(prompt):
    """
    Show a prompt

    :param prompt: Prompt name
    """
    print(cli2.t.y.bold('PATH'))
    print(cli2.t.orange(prompt.path))
    print()

    print(cli2.t.y.bold('CONTENT'))
    print(prompt.parts[0])


@cli.cmd
def render(prompt, **context):
    """
    Render a prompt with a given template context.

    :param prompt: Prompt name
    :param context: Context variables.
    """
    print(cli2.t.y.bold('PATH'))
    print(cli2.t.orange(prompt.path))
    print()

    print(cli2.t.y.bold('OUTPUT'))
    print(prompt.render())


@cli.cmd
def parser(name):
    """
    Show a parser

    :param name: Parser name to display
    """
    plugins = importlib.metadata.entry_points(
        name=name,
        group=prompt2.Parser.entry_point,
    )
    if not plugins:
        print(cli2.t.red.bold('PARSER NOT FOUND') + f': {name}\n')
        print(cli2.t.green.bold('AVAILABLE') + ':')
        return [*parsers().keys()]
    plugin = [*plugins][0]
    parser = plugin.load()

    path = inspect.getfile(parser)
    print(cli2.t.y.bold('PATH'))
    print(cli2.t.o(path))
    print()

    source = inspect.getsource(parser)
    print(cli2.highlight(source, 'Python'))


@cli.cmd
def parsers():
    """
    List registered parsers
    """
    return {
        plugin.name: plugin.value
        for plugin in importlib.metadata.entry_points(
            group=prompt2.Parser.entry_point,
        )
    }


@cli.cmd
def messages(prompt, parser=None, model=None, **context):
    """
    Render prompt messages with a given template context.

    :param name: Prompt name
    :param parser: Parser name if any
    :param model: Model name to use, if any
    :param context: Context variables.
    """
    messages = prompt.messages()
    if parser:
        messages = parser.messages(messages)

    print(cli2.t.y.bold('PATH'))
    print(cli2.t.orange(prompt.path))
    print()
    print(cli2.t.y.bold('OUTPUT'))
    return messages


@cli.cmd
async def send(prompt, parser=None, model=None, **context):
    """
    Send a prompt, rendered with a context, on a model with a parser.

    :param prompt: Prompt name
    :param parser: Parser name if any
    :param model: Model name to use, if any
    :param context: Context variables
    """
    return await model(prompt, parser)
